// interface for channel-wise sample-by-sample processing
#include <math.h>
#include <stdlib.h>

#define CARRIERFREQ 8000 // carrier frequency in Hz
#define MODULATE 0
#define SINGLECHANNEL 0

// if INC_TRANS_CHANNEL==0 then communication channel is not simulated
// if INC_TRANS_CHANNEL==1 a simple constant gain is simulated
// if INC_TRANS_CHANNEL==2 a time-varying constant gain is simulated
#define INC_TRANS_CHANNEL 0

// number of samples per symbol 
#define  N 16 // defines the symbol rate

// modulation scheme used 4-QAM, 16-QAM, 64-QAM, 256-QAM, or 1024-QAM   
#define  QAM 4  

// Number of bits to generate
// this of course depends on the modulation scheme used
#if QAM==4
	#define NBits 2
	#define POW  1
#endif	
#if QAM==16
	#define NBits 4
	#define POW  2
#endif	
#if QAM==64
	#define NBits 6
	#define POW  3
#endif
#if QAM==256
	#define NBits 8
	#define POW  4
#endif
#if QAM==1024
	#define NBits 10
	#define POW  5
#endif

// amplitude of the sync signal 
int sync=30000;

// Symbol amplitude
float A=10000; 

// filter designed in MATLAB with rcosfir() function
// Design parameters: Roll-off=0.5, N_T=1, RATE=16 
#define RCosFilt 33
float G[]={0,
0.0536,
0.1159,
0.1859,
0.2625,
0.3441,
0.4290,
0.5151,
0.6002,
0.6822,
0.7587,
0.8278,
0.8872,
0.9354,
0.9709,
0.9927,
1.0000,
0.9927,
0.9709,
0.9354,
0.8872,
0.8278,
0.7587,
0.6822,
0.6002,
0.5151,
0.4290,
0.3441,
0.2625,
0.1859,
0.1159,
0.0536,
0};

// counter of output samples (samples per symbol)
int counter=0;

// time index for the carrier generation
int n=0;

// time index for the time varying channel
int CHn=0;




float pi=3.14159265358979;


// delay line for the FIR filters  
float  z_I[RCosFilt], z_Q[RCosFilt];

// pointers to the head of the FIR delay lines 
int curr_I=-1, curr_Q=-1;  


// random data bits (in words) before mapping
int b[NBits];

// inphase and quadrature 
float I=0, Q=0;
float OutI=0, OutQ=0; 
// carriers
float SinCarrier=0, CosCarrier=0;



void init()
{
	int i=1;
	// clearing the delays for the FIR filters
	for(i=0; i<(2*N-1); i++)
	{
		z_I[i]=0;
		z_Q[i]=0;
	}	
}

// This function implements FIR filter
// function interface
//   int *curr : a pointer to the head of the circular delay line 
//   float input : current input sample
//   int ntaps  : number of taps in the filter
//   float h[] : filter IR
//   float z[] : delay line of the filter
float fir(int *curr, float input, int ntaps, const float h[], float z[])       
{
    int i=0, curr_i=0;
    float Rez=0;
    
    // circularly increment the pointer to the head 
    (*curr)++;
    // check if we have reached the end of the delay line
    if (*curr==ntaps)
    	*curr=0;
    // put a new sample in the delay line 
    z[*curr]=input;
    
    // implement the convolution sum
    curr_i=*curr;	 	 
    Rez = 0;
    for (i = 0; i < ntaps; i++) 
    {
       // here we read the elements from the delay line backwards
       Rez += h[i] * z[curr_i];
       // decrement the index of the input circularly
       --curr_i;
       // check if the beginning of the delay line is reached
       if (curr_i==-1)
          curr_i=ntaps-1;
        
    }

    return Rez;
}

// QAM mapping of the bits into symbols
void QAMmap()
{
	int i=0;
	
	I=0;
	Q=0;
	for(i=0; i<POW; i++) 
	{
		I=(float)b[2*i]+I*0.5;
		Q=(float)b[2*i+1]+Q*0.5;
	}	
	I=I*A;
	Q=Q*A;
}


// the very simple channel (c+jd)delta[n];
// i.e., a constant (complex) multiplier
void channel()
{
	// real and imaginary part of the channel 
	float c=0, d=0;
	
	// very simple channel
	if (INC_TRANS_CHANNEL==1)
	{ 
		c=1.1;
		d=-0.45;
	}	
	// time-varying channel
	if (INC_TRANS_CHANNEL==2)
	{
		// channel varies at a rate of 1Hz
		// only the phase is affected!!!!
		c=1*cosf(2*pi*CHn*1/32000);
		d=1*sinf(2*pi*CHn*1/32000);
	}
	CHn++;
	if (CHn==32000)
		CHn=0;
		
	I=I*c-Q*d;
	Q=I*d+Q*c;
}

// 4-QAM modulation
void QAMmod()
{
    int i=0;
    // generate random bits
    for(i=0; i<NBits; i++)
        b[i]=((rand()-16378)>0)?1:-1;
        
    // Mapping the bits to QAM 
    QAMmap();
    // include channel if needed 
    #if INC_TRANS_CHANNEL!=0
    	channel();
    #endif	
}

// carrier generation
void carrier()
{
	CosCarrier=2*cosf(2*pi*CARRIERFREQ*n/32000);
	SinCarrier=2*sinf(2*pi*CARRIERFREQ*n/32000);
	n++;
	if (n==32000)
		n=0;

}

/***********************************************/
/**  Modulator                                 */
/***********************************************/

void qam()
{

  // carrier generation 
  #if MODULATE==1
  	carrier();	
  #endif
  
  
  if (counter<N)
  {
  	// wait for the next symbol period
 	
 	
 	//Output Quadrature and In-phase components 
 	//  PUT YOUR MODIFICATIONS HERE -----
 	OutI=0;
  	OutQ=0;
  	//-----------------------------------
  	
  	
  	
   	counter++;
  }
  else
  {
	// The new symbol arrives
	// QAM pulse modulator
	QAMmod();


	//Output Quadrature and In-phase components 
	//  PUT YOUR MODIFICATIONS HERE  ---------
	OutI= (float)I;
   	OutQ= (float)Q;
   	//----------------------------------------

   	
	// set counter two 1, since one sample has been 
	// has been already sent to the output 
	counter=1;
	
	// inverting the sync sign on the symbol boundary
 	sync=-1*sync;
  }
  
  #if MODULATE==1
 		OutI=OutI*CosCarrier;
 		OutQ=OutQ*SinCarrier;
  #endif
  #if SINGLECHANNEL==1
  		// reducing the Vpp of the resulting 
  		// waveform
 		OutQ =0.5*(OutQ-OutI);
 		// the other channel is used for sync
 		OutI= sync;
  #endif 
}

// output channels
int processing_l( int in)
{ 
   return (int)OutI;
}

int processing_r( int in)
{ 
   return (int)OutQ;	
}
